-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "SF_MixSound"

-- **************************************************
-- General information about this script
-- **************************************************

SF_MixSound = {}

function SF_MixSound:Name()
	return "MixSound"
end

function SF_MixSound:Version()
	return "BETA 1"
end

function SF_MixSound:Description()
	return "Mix an audio file with the current soundtrack at the current frame."
end

function SF_MixSound:Creator()
	return "Brian Quinn (c) 2006"
end

function SF_MixSound:UILabel()
	return("MixSound")
end
-- **************************************************
-- Recurring values
-- **************************************************

SF_MixSound.path = nil


-- **************************************************
-- The guts of this script
-- **************************************************




function SF_MixSound:IsEnabled(moho)
	return true
end

function SF_MixSound:Run(moho)

	local oldSoundtrack = moho.document:Soundtrack()
	local x,fileExtension = nil,nil
	local frame = moho.frame
	if (oldSoundtrack == "") then 
		LM.GUI.Alert(LM.GUI.ALERT_INFO, "No Soundtrack loaded!", nil, nil,"EXIT",nil,nil)
		return
	end
	
	local doIt = LM.GUI.Alert(LM.GUI.ALERT_INFO, "Choose a .WAV file. The file will be padded with silence so that", "it begins at the current frame,saved as a new file in the", "same directory, and made the soundtrack.", "OK", nil, "Cancel")
	if doIt == 2 then 
		return 
	end
	
	local validAudioFile = false
	while not validAudioFile do
		local path = LM.GUI.OpenFile("Select Audio File (.wav or .lwv)")
		if (path == "") then
			return
		end
		fileExtension = string.lower(string.sub(path , -4 , -1))
		if (fileExtension == ".wav") or (fileExtension == ".lwv") then
			validAudioFile = true
			self.path = path
		else
			local cancel = LM.GUI.Alert(LM.GUI.ALERT_WARNING, "Please choose a .WAV or .LWV file.", nil, nil, "OK", nil, "Cancel")
			if cancel == 2 then return end
		end
	end

	local wavFile = io.open(self.path, "rb")
	local sampleRate, channels, bytesPerSample, audioStart, audioSize, byteOrder, LOF = SF_MixSound.Audio.ValidateWavFile(wavFile, self.path)
--	print ("1st wav",sampleRate, channels, bytesPerSample, audioStart, "as",audioSize, byteOrder, LOF)	
	if sampleRate == nil then SF_MixSound.BadFile(channels) return end
	local sampleBits = (bytesPerSample / channels) * 8		
	local frameSamples = SF_MixSound.Audio.SamplesPerFrame(moho, sampleRate)
	local frameBytes = frameSamples * bytesPerSample
	local framesToPad = frame - 1
	local paddingBytes = frameBytes * framesToPad
	local padByte = string.char(0)
	if sampleBits == 8 then 
		padByte = string.char(128)
	end

	local wavefileLength = wavFile:seek("end")
	wavFile:seek("set", audioStart)
--	print("start loading new sound from file")
	local newSound = {}
--	print ("padding", paddingBytes)
	for x = 1, paddingBytes + 1 do
		newSound[x] = padByte
	end
	
	for x = 1, audioSize - 4 do
		newSound[x + paddingBytes ] = wavFile:read(1)
	end
--	print("end loading")

	wavFile:close()
	
--------------------------------read the soundtrack and mix
--	print("second mix")
	local soundtrackFile = io.open(oldSoundtrack, "rb")
	sampleRate, channels, bytesPerSample, audioStart, audioSize, byteOrder, LOF = SF_MixSound.Audio.ValidateWavFile(soundtrackFile, oldSoundtrack)
		
	if sampleRate == nil then SF_MixSound.BadFile(channels) return end
	sampleBits = (bytesPerSample / channels) * 8		
	frameSamples = SF_MixSound.Audio.SamplesPerFrame(moho, sampleRate)
	frameBytes = frameSamples * bytesPerSample
	
	paddingBytes = 0
	
--	print(" make new header")
--	print ("audioStart = ", audioStart)
	soundtrackFile:seek("set",0)
	local newHeader = soundtrackFile:read(audioStart)
--	local newHeader = SF_MixSound.Audio.NewWavHeader(soundtrackFile, paddingBytes , byteOrder)
--	print(newHeader)
	
	wavefileLength = soundtrackFile:seek("end")
	soundtrackFile:seek("set",audioStart)
	
--	print("start loading soundtrack from file")
	local soundtrackWav = {}
	for x = 1, wavefileLength - audioStart do
		soundtrackWav[x] = soundtrackFile:read(1)
	end
--	print("end loading")
	
--	soundtrackFile:seek("set",audioStart)
--	local soundtrackWav = soundtrackFile:read(wavefileLength - audioStart)
	soundtrackFile:close()
	
--	print(" Mixing...",channels, bytesPerSample, byteOrder)
	local Mixed = SF_MixSound.Mixit(newSound, soundtrackWav, channels, bytesPerSample, byteOrder)
--	print("Mixed...")
	
--	local NewSoundtrackWav  =  newHeader..Mixed
	local soundtrackPath = SF_MixSound.removeFilenameFromPath(oldSoundtrack)
	local soundtrackFilename = SF_MixSound.filenameFromPath(oldSoundtrack)
	local soundtrackExt = SF_MixSound.extFromFilename(soundtrackFilename)
	soundtrackFilename = SF_MixSound.removeExtFromFilename(soundtrackFilename).."Mixed."..soundtrackExt

	soundtrackPath = soundtrackPath..soundtrackFilename
	
	local soundtrackWavFile = io.open(soundtrackPath, "wb")
	
	if soundtrackWavFile == nil then 
		SF_MixSound.BadFile("File already open! Cannot write to PADDED file.")
		soundtrackWavFile:close() 
		return 
	end
	
--	print("write new wav")
	soundtrackWavFile:seek("set",0)
	soundtrackWavFile:write(newHeader)
	soundtrackWavFile:seek("set",audioStart)
	for x = 1, table.getn(Mixed) do
		soundtrackWavFile:write(Mixed[x])
		--soundtrackWavFile:write(soundtrackWav[x])
	end
	soundtrackWavFile:close()
--	soundtrackWavFile:close()

	moho:SetSoundtrack(soundtrackPath)	

	LM.GUI.Alert(LM.GUI.ALERT_INFO, "New File: "..soundtrackPath.." sucessfully created", "in same directory as original audio file","and assigned as new soundtrack","DONE",nil,nil)

	
end 


--------------------------------------------------------------------------------------------
-- the sevenfeetdesign moho scripting toolbox for dummies (like me)
--------------------------------------------------------------------------------------------

-- (c)2005 B. Quinn -- released under GNU GPL License (www.gnu.org)
-- no warranties of any kind throughout the history of the universe.
-- But it seems to work just fine...
-- last revision 4/8/2005 3:15am - Times are Eastern Daylight Time

-- v1.03 04/12/2005 6:21pm
-- v1.04 04/26/2005 7:15pm
-- v1.05 05/03/2006 3:45am

--SF = {} -- Define the great and terrible SF class
SF_MixSound.Audio = {}
SF_MixSound.AudioFilter = {}
SF_BYTE, SF_WORD, SF_DWORD, SF_QUADWORD = 1, 2, 4, 8
SF_Blen, SF_Wlen, SF_DWlen, SF_QWlen = 8, 16, 32, 64
SF_INTEL, SF_MOTOROLA = 1, -1
SF_BYTESPERSAMPLE = {0, 32768}
SF_MAX_VOLUME = {255, 32767}
SF_e =2.718281828459045
SF_LOWPASS, SF_BANDPASS, SF_HIGHPASS = 1, 2, 3

-----------------------------------------------------------------------------------------------------
--mixomatic
-----------------------------------------------------------------------------------------------------

--SF_MixSound.Mixit(newWav, soundtrackWav, channels, bytesPerSample, byteorder)

function  SF_MixSound.Mixit(inList1, inList2, channels, bytesPerSample, byteorder)     
	
	local outString = nil
	local sampleLen = bytesPerSample / channels
	local outList = {}
	local zip = string.char(0)	
	local zip8 = string.char(128)
	for sample = 1, (table.getn(inList2) - sampleLen), sampleLen do
--		thisSample1 = SF_MixSound.Bytes(string.sub(inList1, sample, sample + (sampleLen - 1)), byteorder)
--		thisSample2 = SF_MixSound.Bytes(string.sub(inList2, sample, sample + (sampleLen - 1)), byteorder)
		local outSample = ""
		if sampleLen == 1 then
			if inList1[sample] = nil then 
				inList1[sample] = zip8
			end
			thisSample1 = string.byte(inList1[sample]) - 127
			thisSample2	= string.byte(inList2[sample]) - 127	
			outList[sample] = LM.Clamp(thisSample1 + thisSample2, -127, 127) + 127
		else
			if byteorder == SF_INTEL then
				if ((inList1[sample]) == nil) or ((inList1[sample + 1]) == nil) then
					inList1[sample], inList1[sample + 1] = zip, zip
				end
				thisSample1 = string.byte(inList1[sample]) + (string.byte(inList1[sample + 1]) * 256)
				thisSample2 = string.byte(inList2[sample]) + (string.byte(inList2[sample + 1]) * 256)
				
			else
				if (inList1[sample] == nil) or ((inList1[sample + 1]) == nil) then
					inList1[sample], inList1[sample + 1] = zip, zip
				end
				thisSample1 = string.byte(inList1[sample + 1]) + (string.byte(inList1[sample]) * 256)
				thisSample2 = string.byte(inList2[sample + 1]) + (string.byte(inList2[sample]) * 256)
			end
			
		--	outSample = LM.Clamp((thisSample1 + thisSample2) , -32767, 32767)
		--	outSample = LM.Clamp((thisSample2) , -32767, 32767)
			if (thisSample1 > 32767) then
				thisSample1 = thisSample1 - 65536
			end
			if (thisSample2 > 32767) then
				thisSample2 = thisSample2 - 65536
			end
		--	local outSample = thisSample1 -- + thisSample1
			local outSample = LM.Clamp((thisSample1 + thisSample2) , -32767, 32767)
		--	local outSample = (thisSample1 + thisSample2) 
			if (outSample < 0) then
				outSample = outSample + 65536
			end
			local stripByte = math.floor(outSample / 256)
			local byte1 = string.char(stripByte)
			local byte2 = string.char(outSample - (stripByte * 256))
			if (sample == 1) then
	--			print ("BO:", byteorder, "samples: 1 ",thisSample1," 2 ", thisSample2, "bytes: 1 ", string.byte(byte1), " 2 ", string.byte(byte2), "outsample: ", outSample)
				if byteorder == SF_INTEL then
					outList[sample] = byte2
					outList[sample + 1] = byte1
				else
					outList[sample] = byte1
					outList[sample + 1] = byte2
				end
			else
				if byteorder == SF_INTEL then
					outList[sample] = byte2
					outList[sample + 1] = byte1
				else
					outList[sample] = byte1
					outList[sample + 1] = byte2
				end
			end
		end
	end
--[[		
	if numBytes > 1 then
		for b = numBytes, 1, -1 do
			local stripByte = math.floor(num / (256 ^ b))
	--print("***",b,stripByte)
			local byte = string.char(stripByte)
			num = num - (stripByte * (256 ^ b))
			bytes[b+1] = byte
			--print("D2B", b, "num", num)
		end 
--		print("past loop")
	end
	bytes[1] = string.char(num)
	
	
		
		if sampleLen == 1 then
			thisSample1 = thisSample1 - 127
			thisSample2	= thisSample2 - 127	
		elseif sampleLen == 2 then
			if (thisSample1 > 32768) then
				thisSample1 = thisSample1 - 65536
			end
			if (thisSample2 > 32768) then
				thisSample2 = thisSample2 - 65536
			end
		end
		
		local outSample = (thisSample1 + thisSample2) 
		
		if (outSample < 0) then
			outSample = outSample + 65536
		end
		
		if (sample == 1) then
			print ("outsample is:", outSample)
			outString = SF_MixSound.Decimal2Bytes(outSample, sampleLen, byteOrder)
			print (thisSample1,thisSample2,outSample, "-", outString)
		else
			outString = outString..SF_MixSound.Decimal2Bytes(outSample, sampleLen, byteOrder)
		end
		
	--	if math.mod(sample, 100) == 0 then print(sample) end
	end
--]]		
	return outList
end



--*******************************************************************************************
-- Simple Warning
--*******************************************************************************************

function SF_MixSound.BadFile(type)
	LM.GUI.Alert(LM.GUI.ALERT_WARNING,"Sorry, incorrect file type or format.", "Error type: "..type, nil, "Exit", nil, nil)
end

-----------------------------------------------------------------------------------------------------
-- Find the specified string or pattern starting from the end of the string
-- input search string, search pattern
-----------------------------------------------------------------------------------------------------
function SF_MixSound.findFromEnd(s, pattern)
	local lastChar = string.len(s)
	local i = -1
	position = nil
	while (position == nil) and ((-i) <= string.len(s)) do
		position = string.find(s, pattern, i)
		if (position == lastChar) then position = nil end
		i = i - 1
	end
	return position
end

-----------------------------------------------------------------------------------------------------
-- Remove the filename from a "/" or "\" delimeted path
-----------------------------------------------------------------------------------------------------
function SF_MixSound.removeFilenameFromPath(path)
	local found = SF_MixSound.findFromEnd(path,"[\\/]")
	if found then 
		return string.sub(path ,1, found)
	end
end

-----------------------------------------------------------------------------------------------------
-- Extract the filename from a "/" or "\" delimeted path
-----------------------------------------------------------------------------------------------------
function SF_MixSound.filenameFromPath(path)
	local found = SF_MixSound.findFromEnd(path,"[\\/]")
	if found then 
		return string.sub(path ,found + 1, string.len(path))
	end
end

-----------------------------------------------------------------------------------------------------
-- Extract the file extension (minus the ".") from a filename
-----------------------------------------------------------------------------------------------------
function SF_MixSound.extFromFilename(filename)
	local found = SF_MixSound.findFromEnd(filename,"[.]")
	if found then 
		return string.lower(string.sub(filename ,found + 1, string.len(filename)))
	end
end

-----------------------------------------------------------------------------------------------------
-- Given a filename as input, return just the filename without a file extension
-----------------------------------------------------------------------------------------------------
function SF_MixSound.removeExtFromFilename(filename)
	local found = SF_MixSound.findFromEnd(filename,"[.]")
	if found then 
		return string.sub(filename, 1, found - 1)
	end
end

-----------------------------------------------------------------------------------------------------
-- Deterine the system operating system
-- Return "win" or "unix" (as strings)
-----------------------------------------------------------------------------------------------------

function SF_MixSound.getOS()
	local opSys = string.lower(string.sub(os.getenv("OS"), 1, 3))
	if opSys == "win" then
		return "win"
	else
		returnb "unix"

	end
end

-----------------------------------------------------------------------------------------------------
-- Just so I remember    -    this one not under my (c)
-----------------------------------------------------------------------------------------------------
function SF_MixSound.createFileList(path, fileFilter)
	
	local cmd = ""
	local fileList = "mohoflst.dat"
	local batchScript = ""
	if(SF_MixSound.getOS()=="win") then
		local drive = string.sub(path, 0 , string.find(path, ":")-1 )
		batchScript = "C:\\mohoimpd.bat"
		cmd = cmd .. drive..":\n"
		cmd = cmd .. "cd \""..path.."\"\n"		
		cmd = cmd .. "DIR /B /OE "..fileFilter.."> "..fileList
	else	--unix (osx or linux)
		batchScript = "~/mohoimpd"
		cmd = cmd .. "cd \""..path.."\"\n"
		cmd = cmd .. "ls -1 "..fileFilter..">"..fileList
	end
	local f = io.open (batchScript, "w")
	if (f == nil) then
		SF_MixSound.BadFile("file write error")
		return
	end
	
	f:write(cmd)
	f:close()

	local result = os.execute ( batchScript ) --no chmod??
	os.remove (batchScript)
	return path..fileList
end
-----------------------------------------------------------------------------------------------------
--Create a list of subdirectories in the chosen directory. Maybe for parsing through
-- a whole directory tree for imports.
-----------------------------------------------------------------------------------------------------
function SF_MixSound.createSubDirList(path)
	local platform = SF_MixSound.getOS()
	local cmd = ""
	local subDirList = "mohodlst.dat"
	local batchScript = ""
	if(self.platform=="win") then
		local drive = string.sub(path, 0 , string.find(path, ":")-1 )
		batchScript = "C:\\mohoimpd.bat"
		cmd = cmd .. drive..":\n"
		cmd = cmd .. "cd \""..path.."\"\n"		
		cmd = cmd .. "DIR /B /AD /OE > "..subDirList
	else	--unix (osx or linux)
		batchScript = "~/mohoimpd"
		cmd = cmd .. "cd \""..path.."\"\n"
		cmd = cmd .. "ls -d1>"..subDirList
	end
	local f = io.open (batchScript, "w")
	if (f == nil) then
		SF_MixSound.BadFile("file write error")
		return
	end
	
	f:write(cmd)
	f:close()

	local result = os.execute ( batchScript ) --no chmod??
	os.remove (batchScript)
	return path..subDirList
end


-----------------------------------------------------------------------------------------------------
-- Return Decimal conversion of string as address, Intel byte order
-----------------------------------------------------------------------------------------------------	

function SF_MixSound.Intel(s)
	local address = 0
	for b = 0, string.len(s) - 1 do
		local byte = string.byte(s, b+1)
		byte = byte * (256 ^ b)
		address = address + byte
	end 
	return address
end

-----------------------------------------------------------------------------------------------------
-- Return Decimal conversion of number as byte string
-----------------------------------------------------------------------------------------------------	

function SF_MixSound.Decimal2Bytes(number, numBytes, byteOrder)
	if byteOrder == nil then byteOrder = SF_INTEL end
	local address = ""
	local bytes = {}
	local num = number
	if numBytes > 1 then
		for b = numBytes, 1, -1 do
			local stripByte = math.floor(num / (256 ^ b))
	--print("***",b,stripByte)
			local byte = string.char(stripByte)
			num = num - (stripByte * (256 ^ b))
			bytes[b+1] = byte
			--print("D2B", b, "num", num)
		end 
--		print("past loop")
	end
	bytes[1] = string.char(num)
	if byteOrder == SF_MOTOROLA then
		for b = numBytes, 1, byteOrder do
			address = address..bytes[b]
		end
	else
		for b = 1, numBytes, byteOrder do
			address = address..bytes[b]
		end
	end
	--print("number", number, "conv", SF_MixSound.Bytes(address, byteOrder))
	return address
end

-----------------------------------------------------------------------------------------------------
-- Return Decimal conversion of string as bytes,interpreted as an unsigned integer
-- Intel or Motorola byte order
-----------------------------------------------------------------------------------------------------	

function SF_MixSound.Bytes(s, byteOrder)
	if byteOrder == nil then byteOrder = SF_INTEL end
	local decimal = 0
	if byteOrder == SF_MOTOROLA then
		for b = string.len(s) - 1, 0, byteOrder do
			local byte = string.byte(s, b+1)
			byte = byte * (256 ^ b)
			decimal = decimal + byte
		end 
	else
		for b = 0, string.len(s) - 1, byteOrder do
			local byte = string.byte(s, b+1)
			byte = byte * (256 ^ b)
			decimal = decimal + byte
		end	
	end
	return decimal
end

-----------------------------------------------------------------------------------------------------
-- Read a number of bytes from a file at a specific location in the file
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Read(file, ptr, size)
	file:seek("set", ptr)
	local byteString = file:read(size)
	return byteString
end

-----------------------------------------------------------------------------------------------------
-- Check if an audio file is a type that can be parsed accurately in Lua. If so, return the
-- files specs, otherwise display an error and return nil.
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.ValidateWavFile(wavFile, path)
	
	local ptrRIFF, ptrLOF, ptrWAVE, ptrSubChunk1Size, ptrAudioFmt, ptrAudioCh, ptrSampleRate, ptrBlockAlign = 
		0, 4, 8, 16, 20, 22, 24, 32 
	local offsetDATA = 20
	local offsetSubChunk2Size = 24
	local offsetAudio = 28 --subchunk1size + offsetAudio is start of data
	local uncompressed = 1
	
	ext = SF_MixSound.extFromFilename(path)
	if ext ~= "lwv" and ext ~= "wav" then return nil, "Bad file extension" end
	
	
	local RIFF = string.upper(SF_MixSound.Read(wavFile, ptrRIFF, SF_DWORD))
	if RIFF ~= "RIFF" and RIFF ~="RIFX" then return nil, "Not valid file" end
	local byteOrder = SF_INTEL
	if RIFF == "RIFX" then byteOrder = SF_MOTOROLA end
		
	if string.upper(SF_MixSound.Read(wavFile, ptrWAVE, SF_DWORD)) ~= "WAVE" then return nil, "Not wav file" end
	local LOF = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrLOF, SF_DWORD), byteOrder)
	
	local subChunk1Size = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrSubChunk1Size, SF_DWORD), byteOrder)
	local DATAtag = string.upper(SF_MixSound.Read(wavFile, subChunk1Size + offsetDATA, SF_DWORD))
--	print (DATAtag)
	if DATAtag == "FACT" then
		offsetDATA = offsetDATA + 12
		offsetSubChunk2Size = offsetSubChunk2Size + 12
		offsetAudio = offsetAudio + 12
		DATAtag = string.upper(SF_MixSound.Read(wavFile, subChunk1Size + offsetDATA, SF_DWORD))
	end
--	print (DATAtag)
	if DATAtag ~= "DATA" then return nil, "Incorrect DATA tag" end
	
	wavFile:seek("set", subChunk1Size + offsetSubChunk2Size)
	local audioSize = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, subChunk1Size + offsetSubChunk2Size, SF_DWORD), byteOrder)
	local audioStart = subChunk1Size + offsetAudio
	
	local audioFormat = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrAudioFmt, SF_WORD), byteOrder)
	if audioFormat ~= 1 then return nil, "Must be uncompressed" end
	
	local channels = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrAudioCh, SF_WORD), byteOrder)
	if (channels <1) and (channels >2) then return nil,"Must be Mono or Stereo only" end
	
	local sampleRate = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrSampleRate, SF_DWORD), byteOrder)
	local bytesPerSample = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrBlockAlign, SF_WORD), byteOrder)
		
	return sampleRate, channels, bytesPerSample, audioStart, audioSize, byteOrder,LOF
end

-----------------------------------------------------------------------------------------------------
-- make a new size wave file header
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.NewWavHeader(wavFile, bytesAdded, byteOrder)
	if byteOrder == nil then byteOrder = SF_INTEL end
	local ptrRIFF, ptrLOF, ptrWAVE, ptrSubChunk1Size, ptrAudioFmt, ptrAudioCh, ptrSampleRate, ptrBlockAlign = 
		0, 4, 8, 16, 20, 22, 24, 32 
	local offsetDATA = 20
	local offsetSubChunk2Size = 24
	local offsetAudio = 28 --subchunk1size + offsetAudio is start of data
	local uncompressed = 1

	local LOF = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrLOF, SF_DWORD), byteOrder)
	local newLOF = LOF + bytesAdded
--	print("oldLOF-newLOF", LOF, newLOF)
	local subChunk1Size = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrSubChunk1Size, SF_DWORD), byteOrder)

	local DATAtag = string.upper(SF_MixSound.Read(wavFile, subChunk1Size + offsetDATA, SF_DWORD))
	if DATAtag == "FACT" then
		offsetDATA = offsetDATA + 12
		offsetSubChunk2Size = offsetSubChunk2Size + 12
		offsetAudio = offsetAudio + 12
		DATAtag = string.upper(SF_MixSound.Read(wavFile, subChunk1Size + offsetDATA, SF_DWORD))
	end

	
	local ptrAudioSize = subChunk1Size + offsetSubChunk2Size
	local audioSize = SF_MixSound.Bytes(SF_MixSound.Read(wavFile, ptrAudioSize, SF_DWORD), byteOrder)
	local newAudioSize = audioSize + bytesAdded
--	print ("audioSize-newAudioSize", audioSize, newAudioSize)
--	print(string.format("Hex Size LOF  %08x  %08x", LOF, newLOF))
--	print(string.format("Hex Size Audio%08x  %08x", audioSize, newAudioSize))	
	local audioStart = subChunk1Size + offsetAudio
	local oldHeader = SF_MixSound.Read(wavFile, 0, audioStart - 4)
--	print ("audiostart", audioStart, string.sub(oldHeader, -4))
	local newHeader = string.sub(oldHeader, 1, 4)
	local newLOFBytes = SF_MixSound.Decimal2Bytes(newLOF, SF_DWORD, byteOrder)
	local newAudioSizeBytes = SF_MixSound.Decimal2Bytes(newAudioSize, SF_DWORD, byteOrder)
--	for x = 1,4 do
--		print (string.format("%02x  %02x", string.byte(string.sub(newLOFBytes,x,x)), string.byte(string.sub(newAudioSizeBytes,x,x))))
--	end
	newHeader = newHeader..newLOFBytes
	newHeader = newHeader..string.sub(oldHeader, 9)
	newHeader = newHeader..newAudioSizeBytes
	
	return newHeader
end

-----------------------------------------------------------------------------------------------------
-- Return the number of bytes and samples in a frames worth of audio
-----------------------------------------------------------------------------------------------------	
function SF_MixSound.Audio.BytesPerFrame(moho, sampleRate, bytesPerSample)
	local fps = moho.document:Fps()	
	local frameSamples = sampleRate / fps
	local frameBytes = frameSamples * bytesPerSample
	return frameBytes
end

-----------------------------------------------------------------------------------------------------
-- Return the number of samples in a frames worth of audio
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.SamplesPerFrame(moho, sampleRate)
	local fps = moho.document:Fps()
	local frameSamples = sampleRate / fps
	return frameSamples
end

-----------------------------------------------------------------------------------------------------
-- Read a frames worth of raw audio data and return it as a string.
-----------------------------------------------------------------------------------------------------
function SF_MixSound.GetAudioFrame(file, frame, stepSize, frameBytes, audioStart)
	local block = SF_MixSound.Read(file, ((frame - 1) * frameBytes) + audioStart, frameBytes * stepSize)
--	print (string.len(block), "block length")
	return block
end

-----------------------------------------------------------------------------------------------------
-- Convert a frames worth of audio (as a single string) into a table of sample values
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.AudioFrame2Sample(moho, block, channels, sampleRate, bytesPerSample, byteOrder)
	local bytePtr = 1
	local sampleSize = bytesPerSample / channels
	local frameBytes = string.len(block)
	local fps = moho.document:Fps()
	local frameSamples = (sampleRate / fps) / channels
	sampleList = {}
	for sample =  1, frameSamples do
		sampleList[sample] = {}
		for channel = 1, channels do
			local sampleVal = string.sub(block, bytePtr, bytePtr + (sampleSize - 1))
			sampleVal = SF_MixSound.Bytes(sampleVal, byteOrder) - SF_BYTESPERSAMPLE[sampleSize]
		
			table.insert(sampleList[sample], channel, sampleVal)
		--	if sample > 6 then print(sampleList[sample][channel], sampleList[sample-5][channel]) end
			bytePtr = bytePtr + sampleSize
		end
	end
	return sampleList
end

-----------------------------------------------------------------------------------------------------
-- Hyperbolic Sine function
-----------------------------------------------------------------------------------------------------
function SF_MixSound.sinh(x)
	return 0.5 * (math.exp(x) - math.exp(-x))
end

-----------------------------------------------------------------------------------------------------
-- Hyperbolic Sine function
-----------------------------------------------------------------------------------------------------
function SF_MixSound.ln(x)
	return math.log10(x) / math.log10(SF_e)
end



-----------------------------------------------------------------------------------------------------
-- Take a table of audio samples and return a new table which has been Lowpass filtered
-- at the specified frequency
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.LowpassFilter(audio, sampleRate, bytesPerSample, frequency)
	local r = math.sqrt(2) 
	
	local c = 1 / math.tan(math.pi * frequency / sampleRate)
	local c2 = c * c
	local rc = r * c
	local a1 = 1 / ( 1 + rc + c2)
	local a2 = 2 * a1
	local a3 = a1
	local b1 = 2 * ( 1 - c2) * a1
	local b2 = ( 1 - rc + c2) * a1
	local outputList = {}
	
	local channels = table.getn(audio[1])
	
	for sample = 1, table.getn(audio) do
		outputList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				local in1 = audio[sample][channel]
				local in2 = audio[sample - 1][channel]
				local in3 = audio[sample - 2][channel]
				local out1 = outputList[sample - 1][channel]
				local out2 = outputList[sample - 2][channel]
				filtered = (a1 * in1) + (a2 * in2) + (a3 * in3) - (b1 * out1) - (b2 * out2)
			else 
				filtered = 0
			end
			table.insert(outputList[sample], channel, filtered)
		end
	end
	return outputList
end

-----------------------------------------------------------------------------------------------------
-- Take a table of audio samples and return a new table which has been Highpass filtered
-- at the specified frequency
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.HighpassFilter(audio, sampleRate, bytesPerSample, frequency)
	local r = math.sqrt(2) 
	
	local c = math.tan(math.pi * frequency / sample_rate)
	local c2 = c * c
	local rc = r * c
	local a1 = 1 / ( 1 + rc + c2)
	local a2 = -2 * a1
	local a3 = a1
	local b1 = 2 * (c2 - 1) * a1
	local b2 = ( 1 - rc + c2) * a1
	local outputList = {}
	local channels = table.getn(audio[1])
	
	for sample = 1, table.getn(audio) do
		outputList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				local in1 = audio[sample][channel]
				local in2 = audio[sample - 1][channel]
				local in3 = audio[sample - 2][channel]
				local out1 = outputList[sample - 1][channel]
				local out2 = outputList[sample - 2][channel]
				filtered = (a1 * in1) + (a2 * in2) + (a3 * in3) - (b1 * out1) - (b2 * out2)
			else 
				filtered = 0
			end
			table.insert(outputList[sample], channel, filtered)
		end
	end
	return outputList
end

-----------------------------------------------------------------------------------------------------
-- Normalize an audio sample list. Works on the original table
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.Normalize(inList, bytesPerSample)
	
	local max = 0
	local channels = table.getn(inList[1])
	for sample = 1, table.getn(inList) do
		for channel = 1, channels do
			if math.abs(inList[sample][channel]) > max then
				max = math.abs(inList[sample][channel])
			end
		end
	end
	
	if max >= SF_MAX_VOLUME[channels] then 
		return
	else
		local volume = (SF_MAX_VOLUME[channels] / max) * .98
		for sample = 1, table.getn(inList) do
			for channel = 1, channels do
				inList[sample][channel] = inList[sample][channel] * volume
			end
		end
	end
	return max
end

-----------------------------------------------------------------------------------------------------
-- Compute the peak volume of a given audio sample list
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.VolumePeak(inList, bytesPerSample)
	
	local max = 0
	local channels = table.getn(inList[1])
	for sample = 1, table.getn(inList)do
	--	print ("inmax", max)
		for channel = 1, channels do
			if math.abs(inList[sample][channel]) > max then
				max = math.abs(inList[sample][channel])
			end
		end
	end
--	print ("max", max)
	return max / SF_MAX_VOLUME[channels]
end

-----------------------------------------------------------------------------------------------------
-- Compute the RMS volume of a given audio sample list, with the sample broken up into 
-- separate sections to determine the maximum RMS.
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.VolumeRMS(inList, bytesPerSample, divisions)
	
	if divisions == nil then divisions = 1 end	
	
	local maxVol = 0
	local max = {}
	local listSize = table.getn(inList)
	local chunkSize = table.getn(inList) / divisions
	local chunkPtr = 1
	local samplePtr = 1
	
	local channels = table.getn(inList[1])
--	print("rms", listSize, chunkSize, channels)
	for chunk = 1, divisions do
	--	print("chunk",chunk)
		local maxVol = 0
		for sample = chunkPtr, chunkSize + (chunkPtr - 1) do
		--	print("sample",sample)
			for channel = 1, channels do
				maxVol = math.abs(inList[sample][channel])^2 + maxVol
			end
		end
		table.insert(max, chunk, math.sqrt(maxVol / (chunkSize * channels)))
		chunkPtr = chunkPtr + chunkSize
	end
	maxVol = 0
	for chunk = 1, divisions do
		if max[chunk] > maxVol then 
			maxVol = max[chunk]
		end
	end
	return maxVol / SF_MAX_VOLUME[channels]
end

-----------------------------------------------------------------------------------------------------
-- Compute the average volume of a given audio sample list, with the sample broken up into 
-- separate sections to determine a "maximum" average. Yeah, it sounds wierd but it works.
-----------------------------------------------------------------------------------------------------
function SF_MixSound.Audio.VolumeAvg(inList, bytesPerSample, divisions)
	
	if divisions == nil then divisions = 4 end	
	
	local maxVol = 0
	local max = {}
	local listSize = table.getn(inList)
	local chunkSize = table.getn(inList) / divisions
	local chunkPtr = 1
	local samplePtr = 1
	
	local channels = table.getn(inList[1])
	
	for chunk = 1, divisions do
		local average = 0
		for sample = chunkPtr, chunkSize + (chunkPtr - 1) do
			for channel = 1, channels do
				maxVol = math.abs(inList[sample][channel]) + maxVol
			end
		end
		max[chunk] = maxVol / (chunkSize * channels)
	end
	maxVol = 0
	for chunk = 1, divisions do
		if max[chunk] > maxVol then 
			maxVol = max[chunk]
		end
	end
	return maxVol / SF_MAX_VOLUME[channels]
end


-----------------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------------

--[[    dBgain (used only for peaking and shelving filters)
    
    Q (the EE kind of definition, except for peakingEQ in which A*Q is
        the classic EE Q.  That adjustment in definition was made so that
        a boost of N dB followed by a cut of N dB for identical Q and
        f0/Fs results in a precisely flat unity gain filter or "wire".)
        
     _or_ BW, the bandwidth in octaves (between -3 dB frequencies for BPF
        and notch or between midpoint (dBgain/2) gain frequencies for
        peaking EQ)
        
     _or_ S, a "shelf slope" parameter (for shelving EQ only).  When S = 1,
        the shelf slope is as steep as it can be and remain monotonically
        increasing or decreasing gain with frequency.  The shelf slope, in
        dB/octave, remains proportional to S for all other values for a
        fixed f0/Fs and dBgain.

--]]
 


-----------------------------------------------------------------------------------------------------
-- Filter fomulaes derived from " Cookbook formulae for audio EQ biquad filter coefficients"
-- by Robert Bristow-Johnson  <rbj@audioimagination.com>
-- Most helpful stuff.
--
-- All filters default to sampleRate=44.1 kHz, Q=1, bandWidth=2 octaves, dBgain=6 dB,
-- slope=1 (as appropriate). Frequencies default to 4kHz for High Pass and Shelf,
-- 100 Hz for Low Pass and Shelf, and 1.5 kHz for everything else.
-----------------------------------------------------------------------------------------------------

-- Q ,the EE kind of definition, except for peakingEQ in which A*Q is the classic EE Q. That 
-- 	adjustment in definition was made so that a boost of N dB followed by a cut of 
-- 	N dB for identical Q and f0/Fs results in a precisely flat unity gain filter. 
--         
-- BW, the bandwidth in octaves (between -3 dB frequencies for BPF	and notch or between 
-- 	midpoint (dBgain/2) gain frequencies for peaking EQ)
--         
-- S, a "shelf slope" parameter (for shelving EQ only). When S = 1, the shelf slope is as 
-- 	steep as it can be and remain monotonically increasing or decreasing gain with 
-- 	frequency. The shelf slope, in dB/octave, remains proportional to S for all other 
-- 	values for a fixed f0/Fs and dBgain.

function SF_MixSound.AudioFilter.LowPass(inList, samplerate, frequency, Q)      

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 100 end
	if Q == nil then Q = 1 end
	
	if frequency == 0 then frequency = .001 end
	if Q == 0 then Q = .001 end
	
	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0/(2*Q)
	
	local b0 =  (1 - cosw0)/2
	local b1 =   1 - cosw0
	local b2 =  (1 - cosw0)/2
	local a0 =   1 + alpha
	local a1 =  -2*cosw0
	local a2 =   1 - alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function  SF_MixSound.AudioFilter.HighPass(inList, samplerate, frequency, Q)     

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 4000 end
	if Q == nil then Q = 1 end
	
	if frequency == 0 then frequency = .001 end
	if Q == 0 then Q = .001 end
	
	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0/(2*Q)
	
	local b0 =  (1 + cosw0)/2
	local b1 = -(1 + cosw0)
	local b2 =  (1 + cosw0)/2
	local a0 =   1 + alpha
	local a1 =  -2*cosw0
	local a2 =   1 - alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function SF_MixSound.AudioFilter.BandPass2(inlist, samplerate, frequency, bandWidth, Q)

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 1500 end
	if bandWidth == nil then bandWidth = 2 end
	if Q == nil then Q = 1 end
	
	if frequency == 0 then frequency = .001 end
	if bandWidth == 0 then bandWidth = .001 end
	if Q == 0 then Q = .001 end
		
	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0 * SF_MixSound.sinh(math.log(2)/2 * bandWidth * w0/sinw0 )
	
	local b0 =   sinw0/2  +   Q*alpha     -- plusses were equals-kinda wierd type thing
	local b1 =   0
	local b2 =  -sinw0/2  +  -Q*alpha
	local a0 =   1 + alpha
	local a1 =  -2*cosw0
	local a2 =   1 - alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function SF_MixSound.AudioFilter.BandPass(inlist, samplerate, frequency, bandWidth, Q)

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 1500 end
	if bandWidth == nil then bandWidth = 2 end
	if Q == nil then Q = 1 end

	if frequency == 0 then frequency = .001 end
	if bandWidth == 0 then bandWidth = .001 end
	if Q == 0 then Q = .001 end
		
	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0 * SF_MixSound.sinh(math.log(2)/2 * bandWidth * w0/sinw0 )
	
	local b0 =   alpha
	local b1 =   0
	local b2 =  -alpha
	local a0 =   1 + alpha
	local a1 =  -2*cosw0
	local a2 =   1 - alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function SF_MixSound.AudioFilter.Notch(inlist, samplerate, frequency, bandWidth, Q) 

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 1500 end
	if bandWidth == nil then bandWidth = 2 end
	if Q == nil then Q = 1 end
	
	if frequency == 0 then frequency = .001 end
	if bandWidth == 0 then bandWidth = .001 end
	if Q == 0 then Q = .001 end

	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0 * SF_MixSound.sinh(math.log(2)/2 * bandWidth * w0/sinw0 )
	
	local b0 =   1
	local b1 =  -2*cosw0
	local b2 =   1
	local a0 =   1 + alpha
	local a1 =  -2*cosw0
	local a2 =   1 - alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function SF_MixSound.AudioFilter.AllPass(inlist, samplerate, frequency, Q) 

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 1500 end
	if Q == nil then Q = 1 end
	
	if frequency == 0 then frequency = .001 end
	if Q == 0 then Q = .001 end


	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0/(2*Q)
	
	local b0 =   1 - alpha
	local b1 =  -2*cosw0
	local b2 =   1 + alpha
	local a0 =   1 + alpha
	local a1 =  -2*cosw0
	local a2 =   1 - alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function SF_MixSound.AudioFilter.PeakingEQ(inlist, samplerate, frequency, Q, dBgain)

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 1500 end
	if Q == nil then Q = 1 end
	if dBgain == nil then dBgain = 6 end
	
	if frequency == 0 then frequency = .001 end
	if Q == 0 then Q = .001 end
	if dBgain == 0 then dBgain = .001 end

	local A  = math.sqrt( 10^(dBgain/20) )
	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0/(2*Q)
	
	local A  = math.sqrt( 10^(dBgain/20) )
	local b0 =   1 + alpha*A
	local b1 =  -2*cosw0
	local b2 =   1 - alpha*A
	local a0 =   1 + alpha/A
	local a1 =  -2*cosw0
	local a2 =   1 - alpha/A
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function SF_MixSound.AudioFilter.LowShelf(inlist, samplerate, frequency, Q, dBgain, slope) 

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 100 end
	if Q == nil then Q = 1 end
	if dBgain == nil then dBgain = 6 end
	if slope == nil then slope = 1 end
	
	if frequency == 0 then frequency = .001 end
	if Q == 0 then Q = .001 end
	if dBgain == 0 then dBgain = .001 end
	if slope == 0 then slope = .001 end

	local A  = math.sqrt( 10^(dBgain/20) )
	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0/2 * sqrt( (A + 1/A)*(1/slope - 1) + 2 )
	
	local b0 =    A*( (A+1) - (A-1)*cosw0 + 2*math.sqrt(A)*alpha )
	local b1 =  2*A*( (A-1) - (A+1)*cosw0                   )
	local b2 =    A*( (A+1) - (A-1)*cosw0 - 2*math.sqrt(A)*alpha )
	local a0 =        (A+1) + (A-1)*cosw0 + 2*math.sqrt(A)*alpha
	local a1 =   -2*( (A-1) + (A+1)*cosw0                   )
	local a2 =        (A+1) + (A-1)*cosw0 - 2*math.sqrt(A)*alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end

function SF_MixSound.AudioFilter.HighShelf(inlist, samplerate, frequency, Q, dBgain, slope)

	if sampleRate == nil then sampleRate = 44100 end
	if frequency == nil then frequency = 4000 end
	if Q == nil then Q = 1 end
	if dBgain == nil then dBgain = 6 end
	if slope == nil then slope = 1 end
	
	if frequency == 0 then frequency = .001 end
	if Q == 0 then Q = .001 end
	if dBgain == 0 then dBgain = .001 end
	if slope == 0 then slope = .001 end

	local A  = math.sqrt( 10^(dBgain/20) )
	local w0 = 2 * math.pi * frequency / sampleRate
	local cosw0 = math.cos(w0)
	local sinw0 = math.sin(w0)
	
	local alpha = sinw0/2 * sqrt( (A + 1/A)*(1/slope - 1) + 2 )
	
	local b0 =    A*( (A+1) + (A-1)*cosw0 + 2*math.sqrt(A)*alpha )
	local b1 = -2*A*( (A-1) + (A+1)*cosw0                   )
	local b2 =    A*( (A+1) + (A-1)*cosw0 - 2*math.sqrt(A)*alpha )
	local a0 =        (A+1) - (A-1)*cosw0 + 2*math.sqrt(A)*alpha
	local a1 =    2*( (A-1) - (A+1)*cosw0                   )
	local a2 =        (A+1) - (A-1)*cosw0 - 2*math.sqrt(A)*alpha
	
	local channels = table.getn(inlist[1])
	
	for sample = 1, table.getn(inList) do
		outList[sample] = {}
		for channel = 1, channels - 1 do
			local filtered = 0
			if sample > 2 then
				filtered =(b0/a0)*inList[sample][channel] + 
				(b1/a0)*inList[sample-1][channel] + 
				(b2/a0)*inList[sample-2][channel] - 
				(a1/a0)*outList[sample-1][channel] - 
				(a2/a0)*outList[sample-2][channel] 
			else 
				filtered = 0
			end
			table.insert(outList[sample], channel, filtered)
		end
	end
	return outList
end






